# SPDX-FileCopyrightText: 2016-2024 Michael Bryan <michaelfbryan@gmail.com> - Ken Mijime <kenaco666@gmail.com>
# SPDX-FileCopyrightText: 2024-present Fabien Hermitte
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import os
import re

from jinja2 import Environment
from jinja2 import FileSystemLoader
from jinja2 import select_autoescape

from foxy_project.changelog.domain_model import Changelog
from foxy_project.changelog.domain_model import PresenterInterface


default_template_dir = os.path.join(os.path.abspath(os.path.dirname(__file__)), "templates")

default_template = {"compact": "compact.jinja2", "lastrelease": "lastrelease.jinja2"}


class MarkdownPresenter(PresenterInterface):  # pylint: disable=too-few-public-methods
    def __init__(self, template=None, issue_url=None):  # noqa: ARG002
        # It is an embedded template
        if template in default_template:
            template_dir = default_template_dir
            template_name = default_template[template]
        # It is a custom template
        else:
            template_dir, template_name = os.path.split(template)

        env = Environment(
            loader=FileSystemLoader(template_dir),
            autoescape=select_autoescape(
                disabled_extensions=("html", "htm", "xml"),
                default_for_string=True,
                default=False,
            ),
        )

        self.template = env.get_template(template_name)

    def present(self, changelog: Changelog) -> str:
        text = self.template.render(changelog=changelog)
        return self._link(changelog.issue_url, changelog.issue_pattern, text)

    @staticmethod
    def _link(url, pattern, text: str) -> str:
        """Replaces all occurrences of pattern in text with markdown links based on url template"""
        if not url or not pattern:
            return text

        def replace(match):
            groups = match.groups()
            if len(groups) == 2:
                matching_text = groups[0]
                ticket_id = groups[1]
            elif len(groups) == 1:
                matching_text = ticket_id = groups[0]
            else:
                raise ValueError("Invalid pattern")
            ticket_url = url.format(id=ticket_id)
            return f"[{matching_text}]({ticket_url})"

        return re.sub(pattern, replace, text)



# SPDX-FileCopyrightText: 2016-2024 Michael Bryan <michaelfbryan@gmail.com> - Ken Mijime <kenaco666@gmail.com>
# SPDX-FileCopyrightText: 2024-present Fabien Hermitte
#
# SPDX-License-Identifier: MIT

from __future__ import annotations


github_issue_pattern = r"(#([\w-]+))"
github_issue_url = "{base_url}/issues/{{id}}"
github_diff_url = "{base_url}/compare/{{previous}}...{{current}}"
github_last_release = "HEAD"

gitlab_issue_pattern = r"(\!([\w-]+))"
gitlab_issue_url = "{base_url}/-/merge_requests/{{id}}"
gitlab_diff_url = "{base_url}/-/compare/{{previous}}...{{current}}"
gitlab_last_release = "master"

default_issue_pattern = github_issue_pattern
default_issue_url = github_issue_url
default_diff_url = github_diff_url
default_last_release = github_last_release


def set_gitlab() -> None:
    global default_issue_pattern  # noqa: PLW0603
    global default_issue_url  # noqa: PLW0603
    global default_diff_url  # noqa: PLW0603
    global default_last_release  # noqa: PLW0603

    default_issue_pattern = gitlab_issue_pattern
    default_issue_url = gitlab_issue_url
    default_diff_url = gitlab_diff_url
    default_last_release = gitlab_last_release


def set_github() -> None:
    global default_issue_pattern  # noqa: PLW0603
    global default_issue_url  # noqa: PLW0603
    global default_diff_url  # noqa: PLW0603
    global default_last_release  # noqa: PLW0603

    default_issue_pattern = github_issue_pattern
    default_issue_url = github_issue_url
    default_diff_url = github_diff_url
    default_last_release = github_last_release



# SPDX-FileCopyrightText: 2016-2024 Michael Bryan <michaelfbryan@gmail.com> - Ken Mijime <kenaco666@gmail.com>
# SPDX-FileCopyrightText: 2024-present Fabien Hermitte
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import logging
import re

from datetime import date
from enum import StrEnum
from hashlib import sha256
from typing import TYPE_CHECKING
from typing import Any
from typing import Optional

from git import Repo
from git import TagReference

from foxy_project.changelog import default_diff_url
from foxy_project.changelog import default_issue_url
from foxy_project.changelog.domain_model import Changelog
from foxy_project.changelog.domain_model import RepositoryInterface
from foxy_project.changelog.domain_model import calendar_nammed_regex
from foxy_project.changelog.domain_model import semver_nammed_regex


if TYPE_CHECKING:
    from git.objects import Commit


class TagPattern(StrEnum):
    SEMVER = "semver"
    CALENDAR = "calendar"


class GitRepository(RepositoryInterface):  # pylint: disable=too-few-public-methods
    def __init__(  # pylint: disable=too-many-arguments
        self,
        repository_path,
        latest_version: str | None = None,
        skip_unreleased: bool = True,
        tag_prefix: str = "",
        tag_pattern: str = "semver",
    ):
        self.repository = Repo(repository_path)
        self.tag_prefix = tag_prefix
        self.tag_pattern = tag_pattern
        self.commit_tags_index = self._init_commit_tags_index(self.repository, self.tag_prefix, self.tag_pattern)
        # in case of defined latest version, unreleased is used as latest release
        self._skip_unreleased = skip_unreleased and not bool(latest_version)
        self._latest_version = latest_version or None

    def generate_changelog(  # pylint: disable=too-many-arguments,too-many-locals
        self,
        title: str = "Changelog",
        description: str = "",
        remote: str = "origin",
        issue_pattern: str | None = None,
        issue_url: str | None = None,
        diff_url: str | None = None,
        starting_commit: str = "",
        stopping_commit: str = "HEAD",
    ) -> Changelog:
        locallogger = logging.getLogger("repository.generate_changelog")
        issue_url = issue_url or self._issue_from_git_remote_url(remote)
        diff_url = diff_url or self._diff_from_git_remote_url(remote)
        changelog = Changelog(title, description, issue_pattern, issue_url, self.tag_prefix, self.tag_pattern)
        if self._repository_is_empty():
            locallogger.info("Repository is empty.")
            return changelog
        iter_rev = self._get_iter_rev(starting_commit, stopping_commit)
        commits = self.repository.iter_commits(
            iter_rev, topo_order=True
        )  # Fixes this bug: https://github.com/KeNaCo/auto-changelog/issues/112
        # Some thoughts here
        #  First we need to check if all commits are "released". If not, we have to create our special "Unreleased"
        #  release. Then we simply iter over all commits, assign them to current release or create new if we find it.
        first_commit = True
        skip = self._skip_unreleased
        locallogger.debug("Start iterating commits")
        for commit in commits:
            sha = commit.hexsha[0:7]
            locallogger.debug("Found commit %s", sha)

            if skip and commit not in self.commit_tags_index:
                locallogger.debug("Skipping unreleased commit %s", sha)
                continue
            skip = False

            if first_commit and commit not in self.commit_tags_index:
                # if no last version specified by the user => consider HEAD
                if not self._latest_version:
                    locallogger.debug("Adding release 'unreleased'")
                    changelog.add_release("Unreleased", sha, date.today(), sha256())
                else:
                    locallogger.debug("Adding release '%s'", self._latest_version)
                    changelog.add_release(self._latest_version, self._latest_version, date.today(), sha256())
            first_commit = False

            if commit in self.commit_tags_index:
                release_attributes = self._extract_release_args(commit, self.commit_tags_index[commit])
                locallogger.debug("Adding release '%s' with attributes %s", release_attributes[0], release_attributes)
                changelog.add_release(*release_attributes)

            note_attributes = self._extract_note_args(commit)
            locallogger.debug("Adding commit %s with attributes %s", sha, note_attributes)
            changelog.add_note(*note_attributes)

        # create the compare url for each release
        releases = changelog.releases
        # we are using len(changelog.releases) - 1 because there is not compare url for the oldest version
        if diff_url is not None:  # if links are off
            for release_index in reversed(range(len(changelog.releases) - 1)):
                releases[release_index].set_compare_url(diff_url, releases[release_index + 1].title)

        # Close the link to the repository
        # If we are not closing it, some references are not cleaned on windows
        self.repository.close()

        return changelog

    def _issue_from_git_remote_url(self, remote: str) -> Optional[str]:
        """Creates issue url with {id} format key"""
        try:
            url = self._remote_url(remote)
            return default_issue_url.format(base_url=url)
        except ValueError as e:
            logging.error("%s. Turning off issue links.", e)
            return None

    def _diff_from_git_remote_url(self, remote: str):
        try:
            url = self._remote_url(remote)
            return default_diff_url.format(base_url=url)
        except ValueError as e:
            logging.error("%s. Turning off compare url links.", e)
            return None

    def _remote_url(self, remote: str) -> str:
        """Extract remote url from remote url"""
        url = self._get_git_url(remote=remote)
        url = GitRepository._sanitize_remote_url(url)
        return url

    @staticmethod
    def _sanitize_remote_url(remote: str) -> str:
        # 'git@github.com:Michael-F-Bryan/auto-changelog.git' -> 'https://github.com/Michael-F-Bryan/auto-changelog'
        # 'https://github.com/Michael-F-Bryan/auto-changelog.git' -> 'https://github.com/Michael-F-Bryan/auto-changelog'
        return re.sub(r"^(https|git|ssh)(:\/\/|@)(.*@)?([^\/:]+)[\/:]([^\/:]+)\/(.+).git$", r"https://\4/\5/\6", remote)

    # This part is hard to mock, separate method is nice approach how to overcome this problem
    def _get_git_url(self, remote: str) -> str:
        remote_config = self.repository.remote(name=remote).config_reader
        # remote url can be in one of this three options
        # Test is the option exits before access it, otherwise the program crashes
        if remote_config.has_option("url"):
            return remote_config.get("url")
        elif remote_config.has_option("pushurl"):
            return remote_config.get("pushurl")
        elif remote_config.has_option("pullurl"):
            return remote_config.get("pullurl")
        else:
            return ""

    def _get_iter_rev(self, starting_commit: str, stopping_commit: str):
        if starting_commit:
            c = self.repository.commit(starting_commit)
            if not c.parents:
                # starting_commit is initial commit,
                # treat as default
                starting_commit = ""
            else:
                # iter_commits iters from the first rev to the second rev,
                # but not contains the second rev.
                # Here we set the second rev to its previous one then the
                # second rev would be included.
                starting_commit = f"{starting_commit}~1"

        iter_rev = f"{stopping_commit}...{starting_commit}" if starting_commit else stopping_commit
        return iter_rev

    def _repository_is_empty(self):
        return not bool(self.repository.references)

    @staticmethod
    def _init_commit_tags_index(
        repo: Repo, tag_prefix: str, tag_pattern: str = "semver"
    ) -> dict[Commit, list[TagReference]]:
        """Create reverse index"""
        reverse_tag_index: dict[Commit, list[TagReference]] = {}
        for tagref in repo.tags:
            tag_name = tagref.name
            commit = tagref.commit

            consider_tag = False

            # consider & remove the prefix if we found one
            if tag_name.startswith(tag_prefix):
                tag_name = tag_name.replace(tag_prefix, "")

                if (tag_pattern == TagPattern.SEMVER and re.fullmatch(semver_nammed_regex, tag_name)) or (  # noqa: SIM114
                    tag_pattern == TagPattern.CALENDAR and re.fullmatch(calendar_nammed_regex, tag_name)
                ):
                    consider_tag = True
                elif re.fullmatch(tag_pattern, tag_name):
                    consider_tag = True
            # good format of the tag => consider it
            if consider_tag:
                if commit not in reverse_tag_index:
                    reverse_tag_index[commit] = []
                reverse_tag_index[commit].append(tagref)
        return reverse_tag_index

    @staticmethod
    def _extract_release_args(commit, tags) -> tuple[str, str, Any, Any]:
        """Extracts arguments for release"""
        title = ", ".join(map(lambda tag: f"{tag.name}", tags))
        date_ = commit.authored_datetime.date()
        sha = commit.hexsha

        # TODO parse message, be carefull about commit message and tags message

        return title, title, date_, sha

    @staticmethod
    def _extract_note_args(commit) -> tuple[str, str, str, str, str, str]:
        """Extracts arguments for release Note from commit"""
        sha = commit.hexsha
        message = commit.message
        type_, scope, description, body, footer = GitRepository._parse_conventional_commit(message)
        return sha, type_, description, scope, body, footer

    @staticmethod
    def _parse_conventional_commit(message: str) -> tuple[str, str, str, str, str]:
        type_ = scope = description = body_footer = body = footer = ""
        # TODO this is less restrictive version of re. I have somewhere more restrictive one, maybe as option?
        match = re.match(r"^(\w+)(\([\w\-.]+\))?!?: (.*)(\n\n[\w\W]*)?$", message.strip())
        if match:
            type_, scope, description, body_footer = match.groups(default="")
        else:
            locallogger = logging.getLogger("repository._parse_conventional_commit")
            locallogger.debug("Commit message did not match expected pattern: %s", message)
        if scope:
            scope = scope[1:-1]
        if body_footer:
            bf_match = re.match(r"^(\n\n[\w\W]+?)?(\n\n([a-zA-Z-]+|BREAKING[- ]CHANGE)(: | #)[\w\W]+)$", body_footer)
            if bf_match:
                result = bf_match.groups(default="")
                body = result[0][2:]
                footer = result[1][2:]
            else:
                body = body_footer[2:]
        return type_, scope, description, body, footer



# SPDX-FileCopyrightText: 2010-2024 Ronny Pfannschmidt <opensource@ronnypfannschmidt.de>
# SPDX-FileCopyrightText: 2024-present Fabien Hermitte
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from foxy_project.changelog.domain_model import calendar_nammed_regex
from foxy_project.changelog.domain_model import semver_nammed_regex
from foxy_project.changelog.repository import GitRepository


if TYPE_CHECKING:
    from typing import Callable
    from typing import Concatenate
    from typing import ParamSpec

    from setuptools_scm.version import ScmVersion

    _P = ParamSpec("_P")


PEP440_FMT = "{guessed}.dev{distance}"
PEP440_NEXT_FMT = "{guessed}"
SEMVER_FMT = "{guessed}-dev{distance}"
SEMVER_NEXT_FMT = "{guessed}"

VERSION_LEN = 3
VERSION_MINOR = 2
VERSION_PATCH = 3
CALENDAR_DECEMBER_MONTH = 12

# Calendar versionning - https://calver.org/


def calendar_conventional_commit_foxy(version: ScmVersion) -> str:
    return _conventional_commit_foxy(
        version, tag_pattern=calendar_nammed_regex, guess_next=guess_next_simple_calendar, fmt=PEP440_FMT
    )


def calendar_conventional_commit_foxy_next(version: ScmVersion) -> str:
    return _conventional_commit_foxy(
        version, tag_pattern=calendar_nammed_regex, guess_next=guess_next_simple_calendar, fmt=PEP440_NEXT_FMT
    )


# Semver versionning - https://semver.org/lang/fr/


def semver_conventional_commit_foxy(version: ScmVersion) -> str:
    return _conventional_commit_foxy(
        version, tag_pattern=semver_nammed_regex, guess_next=guess_next_simple_semver, fmt=SEMVER_FMT
    )


def semver_conventional_commit_next_foxy(version: ScmVersion) -> str:
    return _conventional_commit_foxy(
        version, tag_pattern=semver_nammed_regex, guess_next=guess_next_simple_semver, fmt=SEMVER_NEXT_FMT
    )


# PEP440 versionning - https://peps.python.org/pep-0440/


def pep440_conventional_commit_foxy(version: ScmVersion) -> str:
    return _conventional_commit_foxy(
        version, tag_pattern=semver_nammed_regex, guess_next=guess_next_simple_semver, fmt=PEP440_FMT
    )


def pep440_conventional_commit_next_foxy(version: ScmVersion) -> str:
    return _conventional_commit_foxy(
        version, tag_pattern=semver_nammed_regex, guess_next=guess_next_simple_semver, fmt=PEP440_NEXT_FMT
    )


def _conventional_commit_foxy(
    version: ScmVersion, tag_pattern: str, guess_next: Callable[Concatenate[ScmVersion, _P], str], fmt: str
) -> str:
    if version.exact:
        return version.format_with("{tag}")

    repository = GitRepository(
        Path("."),
        latest_version=None,
        skip_unreleased=False,
        tag_prefix="",
        tag_pattern=tag_pattern,
    )
    changelog = repository.generate_changelog()

    if changelog.releases[0].has_features:
        return version.format_next_version(guess_next, fmt=fmt, retain=VERSION_MINOR)

    return version.format_next_version(guess_next, fmt=fmt, retain=VERSION_PATCH)


def guess_next_simple_calendar(version: ScmVersion, *, retain: int, increment: bool = True) -> str:
    try:
        parts = [int(i) for i in str(version.tag).split(".")[:retain]]
    except ValueError:
        msg: str = f"{version} can't be parsed as numeric version"
        raise ValueError(msg) from None
    while len(parts) < retain:
        parts.append(1)
    if increment:
        parts[-1] += 1
    # Increment the major et reset month to 1 at the end of the year
    if parts[-1] > CALENDAR_DECEMBER_MONTH and retain == VERSION_MINOR:
        parts[-2] += 1
        parts[-1] = 1
    while len(parts) < VERSION_LEN:
        parts.append(1)
    return f"{parts[0]:04}.{parts[1]:02}.{parts[2]:01}"


def guess_next_simple_semver(version: ScmVersion, *, retain: int, increment: bool = True) -> str:
    try:
        parts = [int(i) for i in str(version.tag).split(".")[:retain]]
    except ValueError:
        msg: str = f"{version} can't be parsed as numeric version"
        raise ValueError(msg) from None
    while len(parts) < retain:
        parts.append(0)
    if increment:
        parts[-1] += 1
    while len(parts) < VERSION_LEN:
        parts.append(0)
    return ".".join(str(i) for i in parts)



# SPDX-FileCopyrightText: 2010-2024 Ronny Pfannschmidt <opensource@ronnypfannschmidt.de>
# SPDX-FileCopyrightText: 2024-present Fabien Hermitte
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import TYPE_CHECKING


if TYPE_CHECKING:
    import os


def read_dist_name_from_setup_cfg(
    input_value: str | os.PathLike[str] = "setup.cfg",
) -> str | None:
    # minimal effort to read dist_name off setup.cfg metadata
    import configparser

    parser = configparser.ConfigParser()
    parser.read([input_value], encoding="utf-8")
    return parser.get("metadata", "name", fallback=None)



# SPDX-FileCopyrightText: 2010-2024 Ronny Pfannschmidt <opensource@ronnypfannschmidt.de>
# SPDX-FileCopyrightText: 2024-present Fabien Hermitte
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import logging

from typing import TYPE_CHECKING
from typing import NamedTuple

from foxy_project.config.setuptools import read_dist_name_from_setup_cfg
from foxy_project.config.toml import TOML_RESULT
from foxy_project.config.toml import read_toml_content


if TYPE_CHECKING:
    from pathlib import Path


_ROOT = "root"
PYPROJECT = "pyproject.toml"
FOXY_PROJECT_TOML = "foxy-project.toml"
TOOL_NAME_TOML = "foxy-project"


class PyProjectData(NamedTuple):
    path: Path
    tool_name: str | None
    project: TOML_RESULT
    section: TOML_RESULT

    @property
    def project_name(self) -> str | None:
        return self.project.get("name")

    @property
    def project_description(self) -> str | None:
        return self.project.get("description")


def read_changelog_config(
    path: Path,
    setion_name: str,
    tool_name: str = TOOL_NAME_TOML,
    *,
    use_tool_name: bool = True,
    require_section: bool = True,
) -> PyProjectData:
    defn = read_toml_content(path, None if require_section else {})
    try:
        section = defn.get("tool", {})[tool_name][setion_name] if use_tool_name else defn.get(setion_name, {})
    except LookupError as e:
        error = f"{path} does not contain a tool.{tool_name} section"
        if require_section:
            raise LookupError(error) from e
        # logging.warning("toml section missing %r", error)
        section = {}

    project = defn.get("project", {})
    return PyProjectData(path, tool_name, project, section)


def get_args_for_pyproject(
    pyproject: PyProjectData,
    dist_name: str | None,
    kwargs: TOML_RESULT,
) -> TOML_RESULT:
    """drops problematic details and figures the distribution name"""
    section = pyproject.section.copy()
    kwargs = kwargs.copy()
    if "relative_to" in section:
        relative = section.pop("relative_to")
        logging.warning(
            "%s : at [tool.%s]\n" "ignoring value relative_to=%s" " as its always relative to the config file",
            pyproject.path,
            pyproject.tool_name,
            relative,
        )
    if "dist_name" in section:
        if dist_name is None:
            dist_name = section.pop("dist_name")
        else:
            assert dist_name == section["dist_name"]  # noqa: S101
            section.pop("dist_name")
    if dist_name is None:
        # minimal pep 621 support for figuring the pretend keys
        dist_name = pyproject.project_name
    if dist_name is None:
        dist_name = read_dist_name_from_setup_cfg()
    if _ROOT in kwargs:
        if kwargs[_ROOT] is None:
            kwargs.pop(_ROOT, None)
        elif _ROOT in section:
            if section[_ROOT] != kwargs[_ROOT]:
                logging.warning(f"root {section[_ROOT]} is overridden by the cli arg {kwargs[_ROOT]}")  # noqa: G004
            section.pop(_ROOT, None)
    return {"dist_name": dist_name, **section, **kwargs}



# SPDX-FileCopyrightText: 2016-2024 Michael Bryan <michaelfbryan@gmail.com> - Ken Mijime <kenaco666@gmail.com>
# SPDX-FileCopyrightText: 2010-2024 Ronny Pfannschmidt <opensource@ronnypfannschmidt.de>
# SPDX-FileCopyrightText: 2024-present Fabien Hermitte
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import logging
import os

from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any
from typing import Iterator

import click

from foxy_project.changelog import set_github
from foxy_project.changelog import set_gitlab
from foxy_project.changelog._config import Configuration
from foxy_project.changelog.presenter import MarkdownPresenter
from foxy_project.changelog.presenter import default_template
from foxy_project.changelog.repository import GitRepository
from foxy_project.config.pyproject_reading import FOXY_PROJECT_TOML
from foxy_project.config.pyproject_reading import PYPROJECT


if TYPE_CHECKING:
    from changelog.domain_model import PresenterInterface
    from changelog.domain_model import RepositoryInterface


def validate_template(ctx: Any, param: Any, value: str | None) -> str | None:  # noqa: ARG001
    # Check if an embedded template is passed in parameter or a jinja2 file
    if value is None or (value in default_template or value.endswith(".jinja2")):
        return value

    msg = "Need to pass an embedded template name or a .jinja2 file"
    raise click.BadParameter(msg)


def generate_changelog(
    repository: RepositoryInterface, presenter: PresenterInterface, *args: Any, **kwargs: Any
) -> str:
    """Use-case function coordinates repository and interface"""
    changelog = repository.generate_changelog(*args, **kwargs)
    return presenter.present(changelog)


@click.command(help="Generate a changelog based on the commit history.")
@click.option(
    "-c",
    "--config",
    type=click.Path(exists=True),
    default=None,
    help="path to 'pyproject.toml' with foxy-project config or 'foxy-project.toml' , "
    "default: looked up in the current or parent directories",
)
@click.option(
    "--gitlab",
    help="Set Gitlab Pattern Generation.",
    is_flag=True,
    default=None,
)
@click.option(
    "--github",
    help="Set GitHub Pattern Generation.",
    is_flag=True,
    default=None,
)
@click.option(
    "-p",
    "--path-repo",
    type=click.Path(exists=True),
    default=".",
    help="Path to the repository's root directory [Default: .]",
)
@click.option("-t", "--title", default=None, type=str, help="The changelog's title [Default: Changelog]")
@click.option("-d", "--description", default=None, type=str, help="Your project's description")
@click.option(
    "-o",
    "--output",
    type=click.Path(exists=True),
    default=None,
    help="The place to save the generated changelog [Default: CHANGELOG.md]",
)
@click.option("-r", "--remote", type=str, default=None, help="Specify git remote to use for links")
@click.option("-v", "--latest-version", type=str, default=None, help="use specified version as latest release")
@click.option("-u", "--unreleased", is_flag=True, default=None, help="Include section for unreleased changes")
@click.option(
    "--template",
    callback=validate_template,
    type=str,
    default=None,
    help="specify template to use [compact, lastrelease] or a path to a custom template, default: compact",
)
@click.option(
    "--diff-url", type=str, default=None, help="override url for compares, use {current} and {previous} for tags"
)
@click.option("--issue-url", type=str, default=None, help="Override url for issues, use {id} for issue id")
@click.option(
    "--issue-pattern",
    type=str,
    default=None,
    help="Override regex pattern for issues in commit messages. Should contain two groups, original match and ID used "
    "by issue-url.",
)
@click.option(
    "--tag-pattern",
    type=str,
    default=None,
    help="Specify regex pattern for version tags [semver, calendar, custom-regex]."
    " A custom regex containing one group named 'version' can be specified.",
    show_default=True,
)
@click.option("--tag-prefix", type=str, default=None, help='prefix used in version tags, default: "" ')
@click.option(
    "--stdout",
    is_flag=True,
    default=None,
)
@click.option("--starting-commit", type=str, help="Starting commit to use for changelog generation", default=None)
@click.option("--stopping-commit", type=str, help="Stopping commit to use for changelog generation", default=None)
@click.option(
    "--debug",
    is_flag=True,
    default=False,
    help="set logging level to DEBUG",
)
def changelog(
    config: click.Path | None,
    path_repo: click.Path,
    gitlab: bool | None,
    github: bool | None,
    title: str | None,
    description: str | None,
    output: click.Path | None,
    remote: str | None,
    latest_version: str | None,
    unreleased: bool | None,
    template: str | None,
    diff_url: str | None,
    issue_url: str | None,
    issue_pattern: str | None,
    tag_prefix: str | None,
    stdout: bool | None,
    tag_pattern: str | None,
    starting_commit: str | None,
    stopping_commit: str | None,
    debug: bool | None,
) -> None:
    if debug:
        logging.basicConfig(level=logging.DEBUG)
        logging.debug("Logging level has been set to DEBUG")

    # Priority to find configuration file user configuration -> foxy-project.toml -> pyproject.toml
    config_file = (
        config or _find_file(str(path_repo), name=FOXY_PROJECT_TOML) or _find_file(str(path_repo), name=PYPROJECT)
    )

    configuration = (
        Configuration.from_file(str(config_file), root=os.path.abspath(str(path_repo)))
        if config_file is not None
        else Configuration()
    )
    # Configurations from command line override configuration from file
    configuration = configuration.copy(
        {
            "gitlab": gitlab,
            "github": github,
            "title": title,
            "description": description,
            "output": Path(str(output)) if output is not None else None,
            "remote": remote,
            "latest_version": latest_version,
            "unreleased": unreleased,
            "template": template,
            "diff_url": diff_url,
            "issue_url": issue_url,
            "issue_pattern": issue_pattern,
            "tag_prefix": tag_prefix,
            "stdout": stdout,
            "tag_pattern": tag_pattern,
            "starting_commit": starting_commit,
            "stopping_commit": stopping_commit,
        }
    )

    if configuration.gitlab:
        set_gitlab()

    if configuration.github:
        set_github()

    # Convert the repository name to an absolute path
    repo = os.path.abspath(str(path_repo))

    repository = GitRepository(
        repo,
        latest_version=configuration.latest_version,
        skip_unreleased=not configuration.unreleased,
        tag_prefix=configuration.tag_prefix,
        tag_pattern=configuration.tag_pattern,
    )
    presenter = MarkdownPresenter(template=configuration.template)
    changelog = generate_changelog(
        repository,
        presenter,
        title=configuration.title,
        description=configuration.description,
        remote=configuration.remote,
        issue_pattern=configuration.issue_pattern,
        issue_url=configuration.issue_url,
        diff_url=configuration.diff_url,
        starting_commit=configuration.starting_commit,
        stopping_commit=configuration.stopping_commit,
    )

    if stdout:
        print(changelog)  # noqa: T201
    else:
        _write_changelog(output=configuration.output, changelog=changelog)


PATTERN = "<!-- foxy-changelog-above -->"


def _write_changelog(output: Path, changelog: Any) -> None:
    with open(output, "a+b") as output_file:
        output_file.seek(0)
        lines = [line.decode("utf-8").rstrip() for line in output_file]
        kept_lines: list[str] = []
        is_pattern_found = False
        # Kept the lines after the pattern
        for line in lines:
            if is_pattern_found:
                kept_lines.append(line)
            if PATTERN in line:
                is_pattern_found = True
                kept_lines.append(line)
        # Remove all the content of the file
        output_file.seek(0)
        output_file.truncate()
        # First writes the new generated changelog
        # Second if some lines are kept, writes it back after the generated changelog
        output_file.write(changelog.encode("utf-8"))
        if len(kept_lines) > 0:
            output_file.write(b"\n")
            for line in kept_lines:
                output_file.write(f"{line}\n".encode())


def _find_file(parent: str, name: str) -> str | None:
    for directory in _walk_potential_roots(os.path.abspath(parent)):
        pyproject = os.path.join(directory, name)
        if os.path.isfile(pyproject):
            return pyproject

    return None


def _walk_potential_roots(root: str, *, search_parents: bool = True) -> Iterator[Path]:
    """
    Iterate though a path and each of its parents.
    :param root: File path.
    :param search_parents: If ``False`` the parents are not considered.
    """
    root_path = Path(root)
    yield root_path
    if search_parents:
        yield from root_path.parents



# SPDX-FileCopyrightText: 2010-2024 Ronny Pfannschmidt <opensource@ronnypfannschmidt.de>
# SPDX-FileCopyrightText: 2024-present Fabien Hermitte
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import logging
import sys

from typing import TYPE_CHECKING
from typing import Any
from typing import Callable
from typing import Dict
from typing import TypedDict
from typing import cast


if sys.version_info >= (3, 11):
    from tomllib import loads as load_toml
else:
    from tomli import loads as load_toml

if TYPE_CHECKING:
    from pathlib import Path

    from typing_extensions import TypeAlias

TOML_RESULT: TypeAlias = Dict[str, Any]
TOML_LOADER: TypeAlias = Callable[[str], TOML_RESULT]


def read_toml_content(path: Path, default: TOML_RESULT | None = None) -> TOML_RESULT:
    try:
        data = path.read_text(encoding="utf-8")
    except FileNotFoundError:
        if default is None:
            raise

        logging.debug("%s missing, presuming default %r", path, default)
        return default
    else:
        return load_toml(data)


class _CheatTomlData(TypedDict):
    cheat: dict[str, Any]


def load_toml_or_inline_map(data: str | None) -> dict[str, Any]:
    """
    load toml data - with a special hack if only a inline map is given
    """
    if not data:
        return {}
    if data[0] == "{":
        data = "cheat=" + data
        loaded: _CheatTomlData = cast(_CheatTomlData, load_toml(data))
        return loaded["cheat"]
    return load_toml(data)



# SPDX-FileCopyrightText: 2016-2024 Michael Bryan <michaelfbryan@gmail.com> - Ken Mijime <kenaco666@gmail.com>
# SPDX-FileCopyrightText: 2024-present Fabien Hermitte
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from datetime import date

import pytest

from foxy_project.changelog import default_issue_pattern
from foxy_project.changelog.domain_model import Changelog
from foxy_project.changelog.presenter import MarkdownPresenter


@pytest.fixture(params=["", "Title"])
def title(request):
    return request.param


@pytest.fixture(params=["", "Description"])
def description(request):
    return request.param


# pylint: disable=redefined-outer-name
@pytest.fixture
def empty_changelog(title, description):
    return Changelog(title, description)


@pytest.fixture
def changelog(title, description):
    return Changelog(title, description)


@pytest.fixture
def markdown_presenter():
    return MarkdownPresenter("compact")


def test_markdown_presenter_empty_changelog(empty_changelog, markdown_presenter):
    markdown = markdown_presenter.present(empty_changelog)
    assert empty_changelog.title in markdown
    assert empty_changelog.description in markdown


def test_markdown_presenter_changelog_with_features(changelog, markdown_presenter):
    changelog.add_release("Unreleased", "HEAD", date(2020, 1, 1), None)
    changelog.add_note("", "feat", "description")
    changelog.add_note("", "feat", "description", scope="scope")
    description = f"{changelog.description}\n\n" if changelog.description else ""
    assert_markdown = (
        f"# {changelog.title}\n\n{description}## Unreleased (2020-01-01)\n\n#### New Features\n\n* "
        f"description\n* (scope): description\n"
    )
    markdown = markdown_presenter.present(changelog)
    assert assert_markdown == markdown


def test_markdown_presenter_custom_template_changelog_with_features(changelog, markdown_presenter):
    # scope in bold
    markdown_presenter = MarkdownPresenter("tests/custom_template/custom_template.jinja2")
    changelog.add_release("Unreleased", "HEAD", date(2020, 1, 1), None)
    changelog.add_note("", "feat", "description")
    changelog.add_note("", "feat", "description", scope="scope")
    description = f"{changelog.description}\n\n" if changelog.description else ""
    assert_markdown = (
        f"# {changelog.title}\n\n{description}## Unreleased (2020-01-01)\n\n#### New Features\n\n"
        f"* description\n* **scope**: description\n"
    )
    markdown = markdown_presenter.present(changelog)
    assert assert_markdown == markdown


def test_markdown_presenter_changelog_with_fixes(changelog, markdown_presenter):
    changelog.add_release("Unreleased", "HEAD", date(2020, 1, 1), None)
    changelog.add_note("", "fix", "description")
    changelog.add_note("", "fix", "description", scope="scope")
    description = f"{changelog.description}\n\n" if changelog.description else ""
    assert_markdown = (
        f"# {changelog.title}\n\n{description}## Unreleased (2020-01-01)\n\n#### Fixes\n\n* "
        f"description\n* (scope): description\n"
    )
    markdown = markdown_presenter.present(changelog)
    assert assert_markdown == markdown


def test_markdown_presenter_changelog_custom_template_with_fixes(changelog):
    # scope in bold
    markdown_presenter = MarkdownPresenter("tests/custom_template/custom_template.jinja2")
    changelog.add_release("Unreleased", "HEAD", date(2020, 1, 1), None)
    changelog.add_note("", "fix", "description")
    changelog.add_note("", "fix", "description", scope="scope")
    description = f"{changelog.description}\n\n" if changelog.description else ""
    assert_markdown = (
        f"# {changelog.title}\n\n{description}## Unreleased (2020-01-01)\n"
        f"\n#### Fixes\n\n* description\n* **scope**: description\n"
    )
    markdown = markdown_presenter.present(changelog)
    assert assert_markdown == markdown


@pytest.mark.parametrize(
    "text, expected",
    [
        ("Some text about #42", "Some text about [#42](http://gitlab.com/issues/42)"),
        (
            "Some text with #12 and #42",
            "Some text with [#12](http://gitlab.com/issues/12) and [#42](http://gitlab.com/issues/42)",
        ),
        ("Some text will be unchanged", "Some text will be unchanged"),
        (
            "This is also valid ticket id #ACH-123",
            "This is also valid ticket id [#ACH-123](http://gitlab.com/issues/ACH-123)",
        ),
    ],
)
def test_link_default_match(text, expected, markdown_presenter):
    issue_url = "http://gitlab.com/issues/{id}"
    linked_text = markdown_presenter._link(issue_url, default_issue_pattern, text)  # pylint: disable=protected-access
    assert linked_text == expected


def test_link_default_url(markdown_presenter):
    linked_text = markdown_presenter._link(  # pylint: disable=protected-access
        "", default_issue_pattern, "Some text about #42"
    )
    assert linked_text == "Some text about #42"



# SPDX-FileCopyrightText: 2016-2024 Michael Bryan <michaelfbryan@gmail.com> - Ken Mijime <kenaco666@gmail.com>
# SPDX-FileCopyrightText: 2024-present Fabien Hermitte
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import logging
import os
import subprocess

from datetime import date

import pytest

from click.testing import CliRunner
from foxy_project.__main__ import main


# pylint: disable=redefined-outer-name
@pytest.fixture
def commands():
    return []


@pytest.fixture
def path_project():
    return os.getcwd()


@pytest.fixture
def test_repo(tmp_path, commands):
    cwd = os.getcwd()
    os.chdir(str(tmp_path))
    init_commands = [
        "git init -q",
        "git config user.name 'John Doe'",
        "git config user.email john.doe@email",
        "git config commit.gpgsign false",  # will prevent gpg pass failures
    ]
    for command in init_commands + commands:
        # shell argument fixes error for strings. Details in link below:
        # https://stackoverflow.com/questions/9935151/popen-error-errno-2-no-such-file-or-directory
        subprocess.run(command, shell=True, check=False)  # pylint: disable=subprocess-run-check
    yield str(tmp_path)
    os.chdir(cwd)


@pytest.fixture
def runner():
    return CliRunner(mix_stderr=False)


@pytest.fixture
def changelog_name():
    return "CHANGELOG.md"


# pylint: disable=redefined-outer-name
@pytest.fixture
def open_changelog(test_repo, changelog_name):
    file = None

    def _open_changelog():
        nonlocal file
        file = open(changelog_name, encoding="utf-8")  # pylint: disable=consider-using-with
        return file

    yield _open_changelog

    if file:
        file.close()


def test_help(runner):
    result = runner.invoke(main, ["--help"])
    assert result.exit_code == 0, result.stderr
    assert result.output


@pytest.mark.parametrize(
    "commands",
    [["git remote add origin https://github.com/Michael-F-Bryan/auto-changelog.git"]],
)
def test_empty_repo(runner, open_changelog):
    """JS test notes:
    * js generates H3 instead of H1
    * js contains explanation what will be content of the file - I don't think this is necessary
    * js contains "Generated by auto-changelog" credentials
    """
    result = runner.invoke(main)
    assert result.exit_code == 0, result.stderr
    assert result.output == ""
    changelog = open_changelog().read()
    assert changelog == "# Changelog\n"


@pytest.mark.parametrize(
    "commands",
    [["git remote add origin https://github.com/Michael-F-Bryan/auto-changelog.git"]],
)
def test_option_repo(test_repo, runner, open_changelog):
    result = runner.invoke(main, ["--path-repo", test_repo])
    assert result.exit_code == 0, result.stderr
    assert result.output == ""
    changelog = open_changelog().read()
    assert changelog == "# Changelog\n"


@pytest.mark.parametrize(
    "commands",
    [["git remote add origin https://github.com/Michael-F-Bryan/auto-changelog.git"]],
)
def test_option_title(runner, open_changelog):
    result = runner.invoke(main, ["--title", "Title"])
    assert result.exit_code == 0, result.stderr
    assert result.output == ""
    changelog = open_changelog().read()
    assert changelog == "# Title\n"


@pytest.mark.parametrize(
    "commands",
    [["git remote add origin https://github.com/Michael-F-Bryan/auto-changelog.git"]],
)
def test_option_description(runner, open_changelog):
    result = runner.invoke(main, ["--description", "My description"])
    assert result.exit_code == 0, result.stderr
    assert result.output == ""
    changelog = open_changelog().read()
    assert changelog == "# Changelog\n\nMy description\n"


@pytest.mark.parametrize("changelog_name", ["a.out"])
@pytest.mark.parametrize(
    "commands",
    [["git remote add origin https://github.com/Michael-F-Bryan/auto-changelog.git"]],
)
def test_option_output(runner, open_changelog):
    result = runner.invoke(main, ["--output", "a.out"])
    assert result.exit_code == 0, result.stderr
    assert result.output == ""
    changelog = open_changelog().read()
    assert changelog == "# Changelog\n"


@pytest.mark.parametrize(
    "commands",
    [
        [
            'git commit --allow-empty -q -m "feat: Add file #1"',
            "git remote add upstream git@github.com:Michael-F-Bryan/auto-changelog.git",
            "git remote add origin git@github.com:KeNaCo/auto-changelog.git",
        ]
    ],
)
def test_option_remote(runner, open_changelog):
    """JS test notes:
    * commit not recognized or included - a bug: https://github.com/cookpete/auto-changelog/issues/174
    """
    result = runner.invoke(main, ["--remote", "upstream", "--unreleased"])
    assert result.exit_code == 0, result.stderr
    assert result.output == ""
    changelog = open_changelog().read()
    assert_content = (
        f"# Changelog\n\n## Unreleased ({date.today().strftime('%Y-%m-%d')})\n\n#### New Features\n\n"
        f"* Add file [#1](https://github.com/Michael-F-Bryan/auto-changelog/issues/1)\n"
    )
    assert changelog == assert_content


@pytest.mark.parametrize(
    "commands",
    [
        [
            'git commit --allow-empty -q -m "feat: Add file #1"',
            "git remote add upstream git@github.com:Michael-F-Bryan/auto-changelog.git",
            "git remote add origin git@github.com:KeNaCo/auto-changelog.git",
        ]
    ],
)
def test_option_default_remote(runner, open_changelog):
    result = runner.invoke(main, ["--unreleased"])  # --remote default: origin
    assert result.exit_code == 0, result.stderr
    assert result.output == ""
    changelog = open_changelog().read()
    assert_content = (
        f"# Changelog\n\n## Unreleased ({date.today().strftime('%Y-%m-%d')})\n\n#### New Features\n\n"
        f"* Add file [#1](https://github.com/KeNaCo/auto-changelog/issues/1)\n"
    )
    assert changelog == assert_content


@pytest.mark.parametrize(
    "commands",
    [
        [
            'git commit --allow-empty -q -m "feat: Add file #1"',
            "git remote add upstream git@github.com:Michael-F-Bryan/auto-changelog.git",
        ]
    ],
)
def test_option_default_missing_remote(runner, open_changelog):
    result = runner.invoke(main, ["--unreleased"])  # --remote default: origin
    assert result.exit_code == 0, result.stderr
    assert result.output == ""
    changelog = open_changelog().read()
    assert_content = (
        f"# Changelog\n\n## Unreleased ({date.today().strftime('%Y-%m-%d')})\n\n#### New Features\n\n"
        f"* Add file #1\n"
    )
    assert changelog == assert_content


@pytest.mark.parametrize(
    "commands",
    [
        [
            'git commit --allow-empty -q -m "feat: Add file #1"',
            "git remote add origin https://github.com/Michael-F-Bryan/auto-changelog.git",
        ]
    ],
)
def test_option_latest_version(runner, open_changelog):
    result = runner.invoke(main, ["--latest-version", "1.0.0"])
    assert result.exit_code == 0, result.stderr
    assert result.output == ""
    changelog = open_changelog().read()
    assert_content = (
        f"# Changelog\n\n## 1.0.0 ({date.today().strftime('%Y-%m-%d')})\n\n#### New Features\n\n"
        f"* Add file [#1](https://github.com/Michael-F-Bryan/auto-changelog/issues/1)\n"
    )
    assert changelog == assert_content


@pytest.mark.parametrize(
    "commands",
    [
        [
            'git commit --allow-empty -q -m "feat: Add file #1"',
        ]
    ],
)
def test_option_unreleased(runner, open_changelog):
    result = runner.invoke(main, ["--unreleased"])
    assert result.exit_code == 0, result.stderr
    assert result.output == ""
    changelog = open_changelog().read()
    assert_content = (
        f"# Changelog\n\n## Unreleased ({date.today().strftime('%Y-%m-%d')})\n\n#### New Features\n\n"
        f"* Add file #1\n"
    )
    assert changelog == assert_content


@pytest.mark.parametrize(
    "commands",
    [
        [
            'git commit --allow-empty -q -m "feat: Add file #1"',
            "git remote add origin https://github.com/Michael-F-Bryan/auto-changelog.git",
        ]
    ],
)
def test_option_skipping_unreleased(runner, open_changelog):
    result = runner.invoke(main)
    assert result.exit_code == 0, result.stderr
    assert result.output == ""
    changelog = open_changelog().read()
    assert changelog == "# Changelog\n"
    assert "## Unreleased" not in changelog


@pytest.mark.parametrize(
    "commands",
    [
        [
            'git commit --allow-empty -q -m "feat: Add file #1"',
            "git remote add origin https://github.com/Michael-F-Bryan/auto-changelog.git",
        ]
    ],
)
def test_option_issue_url(runner, open_changelog):
    result = runner.invoke(main, ["--issue-url", "issues.custom.com/{id}", "--unreleased"])
    assert result.exit_code == 0, result.stderr
    assert result.output == ""
    changelog = open_changelog().read()
    assert_content = (
        f"# Changelog\n\n## Unreleased ({date.today().strftime('%Y-%m-%d')})\n\n#### New Features\n\n"
        f"* Add file [#1](issues.custom.com/1)\n"
    )
    assert changelog == assert_content


@pytest.mark.parametrize(
    "commands",
    [
        [
            'git commit --allow-empty -q -m "feat: Add file PRO-1"',
            "git remote add origin https://github.com/Michael-F-Bryan/auto-changelog.git",
        ]
    ],
)
def test_option_issue_pattern(runner, open_changelog):
    result = runner.invoke(
        main,
        ["--issue-pattern", r"([a-zA-Z][a-zA-Z][a-zA-Z]-\d+)", "--issue-url", "issues.custom.com/{id}", "--unreleased"],
    )
    assert result.exit_code == 0, result.stderr
    assert result.output == ""
    changelog = open_changelog().read()
    assert_content = (
        f"# Changelog\n\n## Unreleased ({date.today().strftime('%Y-%m-%d')})\n\n#### New Features\n\n"
        f"* Add file [PRO-1](issues.custom.com/PRO-1)\n"
    )
    assert changelog == assert_content


@pytest.mark.parametrize(
    "commands",
    [
        [
            'git commit --allow-empty -q -m "feat: Add file PRO-1"',
            "git remote add origin https://github.com/Michael-F-Bryan/auto-changelog.git",
        ]
    ],
)
def test_option_invalid_issue_pattern(runner, open_changelog):
    result = runner.invoke(
        main,
        ["--issue-pattern", r" [a-zA-Z][a-zA-Z][a-zA-Z]-\d+", "--issue-url", "issues.custom.com/{id}", "--unreleased"],
    )
    assert result.exit_code != 0, result.stderr


@pytest.mark.parametrize(
    "commands",
    [["git remote add origin https://github.com/Michael-F-Bryan/auto-changelog.git"]],
)
def test_option_stdout(runner, open_changelog):
    result = runner.invoke(main, ["--stdout"])
    assert result.exit_code == 0, result.stderr
    assert result.output == "# Changelog\n\n"


@pytest.mark.parametrize(
    "commands",
    [
        [
            'git commit --allow-empty -q -m "feat: Add file"',
            "git tag custom-tag",
            'git commit --allow-empty -q -m "chore: Change"',
            "git tag 1.0.0",
            'git commit --allow-empty -q -m "chore: Change2"'
            "git tag v2.0.0"
            "git remote add origin https://github.com/Michael-F-Bryan/auto-changelog.git",
        ]
    ],
)
def test_option_tag_pattern(runner, open_changelog):
    """JS test notes:
    * we are using full-match but JS is using match
    """
    result = runner.invoke(main, ["--tag-pattern", r"\d+.\d+.\d+"])
    assert result.exit_code == 0, result.stderr
    changelog = open_changelog().read()
    assert "1.0.0" in changelog
    assert_content = (
        f"# Changelog\n\n## 1.0.0 ({date.today().strftime('%Y-%m-%d')})\n\n#### New Features\n\n"
        f"* Add file\n#### Others\n\n* Change\n"
    )
    assert changelog == assert_content


@pytest.mark.parametrize(
    "commands",
    [
        [
            'git commit --allow-empty -q -m "feat: Add file"',
            "git tag v-something",
            'git commit --allow-empty -q -m "chore: Change"',
            "git tag 1.0.0",
            "git tag v2.0.0",
            'git commit --allow-empty -q -m "chore: Change2"',
            "git tag v3.0.0",
            "git remote add origin https://github.com/Michael-F-Bryan/auto-changelog.git",
        ]
    ],
)
def test_option_tag_prefix(runner, open_changelog):
    """JS test notes:
    * JS version is a mess, reported https://github.com/cookpete/auto-changelog/issues/193
    """
    result = runner.invoke(main, ["--tag-prefix", "v"])
    assert result.exit_code == 0, result.exc_info
    changelog = open_changelog().read()
    assert_content = (
        f"# Changelog\n\n## v3.0.0 ({date.today().strftime('%Y-%m-%d')})\n\n#### Others\n\n* Change2\n\n"
        f"Full set of changes: [`v2.0.0...v3.0.0`]"
        f"(https://github.com/Michael-F-Bryan/auto-changelog/compare/v2.0.0...v3.0.0)\n\n"
        f"## v2.0.0 ({date.today().strftime('%Y-%m-%d')})\n\n#### New Features\n\n* Add file\n"
        f"#### Others\n\n* Change\n"
    )
    assert changelog == assert_content
    assert "1.0.0" not in changelog
    assert "v-something" not in changelog
    assert "v2.0.0" in changelog
    assert "v3.0.0" in changelog


@pytest.mark.parametrize(
    "commands",
    [
        [
            'git commit --allow-empty -q -m "feat: Add file"',
            "git tag release-1",
            'git commit --allow-empty -q -m "chore: Change"',
            "git tag 1",
            "git tag release-1.2.3",
            "git remote add origin https://github.com/Michael-F-Bryan/auto-changelog.git",
        ]
    ],
)
def test_tag_prefix_and_pattern_combination(runner, open_changelog):
    """Combination of prefix and pattern work in a way that first prefix tags are filtered,
    then tag - (minus) prefix part is checked with the pattern.

    JS test notes:
    * includes also release-1.2.3
    """
    result = runner.invoke(main, ["--tag-prefix", "release-", "--tag-pattern", r"\d"])
    assert result.exit_code == 0, result.stderr
    changelog = open_changelog().read()
    assert_content = (
        f"# Changelog\n\n## release-1 ({date.today().strftime('%Y-%m-%d')})\n\n" f"#### New Features\n\n* Add file\n"
    )
    assert changelog == assert_content
    assert "## 1 " not in changelog
    assert "release-1.2.3" not in changelog


@pytest.mark.parametrize(
    "commands",
    [
        [
            'git commit --allow-empty -q -m "feat: Add file PRO-1"',
            'git commit --allow-empty -q -m "fix: Some file fix"',
            "git tag start",
            "git tag 1.0.0",
            "git remote add origin https://github.com/Michael-F-Bryan/auto-changelog.git",
        ]
    ],
)
def test_starting_commit(test_repo, runner, open_changelog):
    """JS test notes:
    * not supported by JS
    * JS have --starting-version [tag] which is more specific variant of this attribute
    * maybe test also different references than tag?
    """
    result = runner.invoke(main, ["--starting-commit", "start"])
    assert result.exit_code == 0, result.stderr
    changelog = open_changelog().read()
    assert_content = f"# Changelog\n\n## 1.0.0 ({date.today().strftime('%Y-%m-%d')})\n\n#### Fixes\n\n* Some file fix\n"
    assert changelog == assert_content
    assert "Add file PRO-1" not in changelog


@pytest.mark.parametrize(
    "commands",
    [
        [
            'git commit --allow-empty -q -m "fix: Some file fix"',
            "git tag 1.0.0",
            "git remote add origin https://github.com/Michael-F-Bryan/auto-changelog.git",
        ]
    ],
)
def test_starting_commit_is_only_commit(runner, open_changelog):
    """JS test notes:
    * not supported by JS (see test above)
    """
    result = runner.invoke(main, ["--starting-commit", "1.0.0"])
    assert result.exit_code == 0, result.stderr
    changelog = open_changelog().read()
    assert_content = f"# Changelog\n\n## 1.0.0 ({date.today().strftime('%Y-%m-%d')})\n\n#### Fixes\n\n* Some file fix\n"
    assert changelog == assert_content


@pytest.mark.parametrize(
    "commands",
    [
        [
            'git commit --allow-empty -q -m "fix: Some file fix"',
            "git tag 1.0.0",
            "git remote add origin https://github.com/Michael-F-Bryan/auto-changelog.git",
        ]
    ],
)
def test_starting_commit_not_exist(test_repo, runner, open_changelog):
    """JS test notes:
    * not supported by JS (see test above)
    """
    result = runner.invoke(main, ["--starting-commit", "nonexist"])
    assert result.exit_code != 0, result.stderr


@pytest.mark.parametrize(
    "commands",
    [
        [
            'git commit --allow-empty -q -m "feat: Add file PRO-1"',
            "git tag stop",
            'git commit --allow-empty -q -m "fix: Some file fix"',
            "git remote add origin https://github.com/Michael-F-Bryan/auto-changelog.git",
        ]
    ],
)
def test_stopping_commit(runner, open_changelog):
    result = runner.invoke(main, ["--stopping-commit", "stop", "--unreleased"])
    assert result.exit_code == 0, result.stderr
    changelog = open_changelog().read()
    assert_content = (
        f"# Changelog\n\n## Unreleased ({date.today().strftime('%Y-%m-%d')})\n\n#### New Features\n\n* Add file PRO-1\n"
    )
    assert changelog == assert_content
    assert "Some file fix" not in changelog


@pytest.mark.parametrize(
    "commands",
    [
        [
            'git commit --allow-empty -q -m "feat: Add file #1\n\n"',
        ]
    ],
)
def test_empty_line_body(test_repo, runner, open_changelog):
    result = runner.invoke(main, ["--unreleased"])
    assert result.exit_code == 0, result.stderr
    assert result.output == ""
    changelog = open_changelog().read()
    assert_content = (
        f"# Changelog\n\n## Unreleased ({date.today().strftime('%Y-%m-%d')})\n\n#### New Features\n\n* Add file #1\n"
    )
    assert changelog == assert_content


@pytest.mark.parametrize(
    "commands",
    [
        [
            'git commit --allow-empty -q -m "feat: Add file #1\n\nBody line"',
        ]
    ],
)
def test_single_line_body(runner, open_changelog):
    result = runner.invoke(main, ["--unreleased"])
    assert result.exit_code == 0, result.stderr
    assert result.output == ""
    changelog = open_changelog().read()
    assert_content = (
        f"# Changelog\n\n## Unreleased ({date.today().strftime('%Y-%m-%d')})\n\n#### New Features\n\n* Add file #1\n"
    )
    assert changelog == assert_content


@pytest.mark.parametrize(
    "commands",
    [["git commit --allow-empty -q -m 'feat: Add file #1\n\nBody line 1\nBody line 2'"]],
)
def test_double_line_body(runner, open_changelog):
    result = runner.invoke(main, ["--unreleased"])
    assert result.exit_code == 0, result.stderr
    assert result.output == ""
    changelog = open_changelog().read()
    print(changelog)
    assert "Add file #1" in changelog


@pytest.mark.parametrize(
    "commands",
    [
        [
            'git commit --allow-empty -q -m "feat: Add file #1\n\nBody line 1\nBody line 2\nBody line 3"',
        ]
    ],
)
def test_triple_line_body(runner, open_changelog):
    result = runner.invoke(main, ["--unreleased"])
    assert result.exit_code == 0, result.stderr
    assert result.output == ""
    changelog = open_changelog().read()
    assert_content = (
        f"# Changelog\n\n## Unreleased ({date.today().strftime('%Y-%m-%d')})\n\n#### New Features\n\n* Add file #1\n"
    )
    assert changelog == assert_content


@pytest.mark.parametrize(
    "commands",
    [
        [
            'git commit --allow-empty -q -m "feat: Add file #1\n\nBody paragraph 1\n\nBody paragraph 2"',
        ]
    ],
)
def test_multi_paragraph_body(runner, open_changelog):
    result = runner.invoke(main, ["--unreleased"])
    assert result.exit_code == 0, result.stderr
    assert result.output == ""
    changelog = open_changelog().read()
    assert_content = (
        f"# Changelog\n\n## Unreleased ({date.today().strftime('%Y-%m-%d')})\n\n#### New Features\n\n* Add file #1\n"
    )
    assert changelog == assert_content


@pytest.mark.parametrize(
    "commands",
    [
        [
            'git commit --allow-empty -q -m "feat: Add file #1\n\nBody line\n\nFooter: first footer"',
        ]
    ],
)
def test_single_line_body_single_footer(runner, open_changelog):
    result = runner.invoke(main, ["--unreleased"])
    assert result.exit_code == 0, result.stderr
    assert result.output == ""
    changelog = open_changelog().read()
    assert_content = (
        f"# Changelog\n\n## Unreleased ({date.today().strftime('%Y-%m-%d')})\n\n#### New Features\n\n* Add file #1\n"
    )
    assert changelog == assert_content


@pytest.mark.parametrize(
    "commands",
    [
        [
            'git commit --allow-empty -q -m "feat: Add file #1\n\nBody line\n\n'
            'Footer: first footer\nFooter: second footer"',
        ]
    ],
)
def test_single_line_body_double_footer(runner, open_changelog):
    result = runner.invoke(main, ["--unreleased"])
    assert result.exit_code == 0, result.stderr
    assert result.output == ""
    changelog = open_changelog().read()
    assert_content = (
        f"# Changelog\n\n## Unreleased ({date.today().strftime('%Y-%m-%d')})\n\n#### New Features\n\n* Add file #1\n"
    )
    assert changelog == assert_content


@pytest.mark.parametrize(
    "commands",
    [
        [
            'git commit --allow-empty -q -m "feat(scope): Add file #1\n\nBody line 1\nBody line 2\nBody line 3"',
        ]
    ],
)
def test_custom_template(path_project, runner, open_changelog):
    custom_template_path = os.path.join(path_project, "tests/custom_template/custom_template.jinja2")
    result = runner.invoke(main, ["--unreleased", "--template", custom_template_path])
    assert result.exit_code == 0, result.stderr
    assert result.output == ""
    changelog = open_changelog().read()
    assert_content = (
        f"# Changelog\n\n## Unreleased ({date.today().strftime('%Y-%m-%d')})\n\n"
        f"#### New Features\n\n* **scope**: Add file #1\n"
    )
    assert changelog == assert_content


def test_custom_template_invalid_file(path_project, runner, caplog):
    result = runner.invoke(main, ["--template", "nonsense"])
    assert result.exit_code != 0, result.stderr


@pytest.mark.parametrize(
    "commands",
    [["git remote add origin https://github.com/Michael-F-Bryan/auto-changelog.git"]],
)
def test_debug(caplog, runner, open_changelog):
    caplog.set_level(logging.DEBUG)
    result = runner.invoke(main, ["--debug"])
    assert result.exit_code == 0, result.stderr
    assert result.output == ""
    assert "Logging level has been set to DEBUG" in caplog.text


@pytest.mark.parametrize(
    "commands",
    [["git remote add origin https://github.com/Michael-F-Bryan/auto-changelog.git"]],
)
def test_no_debug(caplog, runner, open_changelog):
    caplog.set_level(logging.DEBUG)
    result = runner.invoke(main)
    assert result.exit_code == 0, result.stderr
    assert result.output == ""
    assert "Logging level has been set to DEBUG" not in caplog.text



# SPDX-FileCopyrightText: 2010-2024 Ronny Pfannschmidt <opensource@ronnypfannschmidt.de>
# SPDX-FileCopyrightText: 2024-present Fabien Hermitte
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import logging
import os

from pathlib import Path
from typing import Iterator

import click

from setuptools_scm import get_version

from foxy_project.config.pyproject_reading import FOXY_PROJECT_TOML
from foxy_project.config.pyproject_reading import PYPROJECT
from foxy_project.version._config import Configuration
from foxy_project.version.version_scheme import calendar_conventional_commit_foxy_next
from foxy_project.version.version_scheme import pep440_conventional_commit_next_foxy
from foxy_project.version.version_scheme import semver_conventional_commit_next_foxy


@click.command(help="View project's version based on the commit history.")
@click.option(
    "-c",
    "--config",
    type=click.Path(exists=True),
    default=None,
    help="path to 'pyproject.toml' with foxy-project config or 'foxy-project.toml' , "
    "default: looked up in the current or parent directories",
)
@click.option(
    "-p",
    "--path-repo",
    type=click.Path(exists=True),
    default=".",
    help="Path to the repository's root directory [Default: .]",
)
@click.option(
    "--version-scheme",
    type=str,
    default=None,
    help="Configures how the local version number is constructed;"
    "either an entrypoint name or a callable. [Default: semver-conventional-commit-foxy]",
)
@click.option(
    "--local-scheme",
    type=str,
    default=None,
    help="Configures how the local version number is constructed;"
    "either an entrypoint name or a callable. [Default: node-and-date]",
)
@click.option(
    "--version-file",
    type=click.Path(exists=True),
    default=None,
    help="A path to a file that gets replaced with a file containing the current version.",
)
@click.option(
    "--version-file-template",
    type=str,
    default=None,
    help="A new-style format string that is given the current"
    "version as the version keyword argument for formatting.",
)
@click.option(
    "--relative-to",
    type=click.Path(exists=True),
    default=None,
    help="A file/directory from which the root can be resolved.",
)
@click.option(
    "--tag-regex",
    type=str,
    default=None,
    help="A Python regex string to extract the version part from any SCM tag."
    "The regex needs to contain either a single match group, or a group named version,"
    "that captures the actual version information.",
)
@click.option(
    "--parentdir-prefix-version",
    type=str,
    default=None,
    help="If the normal methods for detecting the version (SCM version, sdist metadata) fail,"
    "and the parent directory name starts with parentdir_prefix_version,"
    "then this prefix is stripped and the rest of the parent directory name"
    "is matched with tag_regex to get a version string.",
)
@click.option(
    "--fallback-version",
    type=str,
    default=None,
    help="A version string that will be used if no other method for detecting the version worked"
    "(e.g., when using a tarball with no metadata).",
)
@click.option(
    "--next",
    is_flag=True,
    default=False,
    help="Use the next possible version instead of the current one.",
)
@click.option(
    "--no-print",
    is_flag=True,
    default=False,
    help="Deactivate the print of the version.",
)
@click.option(
    "--no-version-file",
    is_flag=True,
    default=False,
    help="Deactivate the generation of the version file.",
)
@click.option(
    "--debug",
    is_flag=True,
    default=False,
    help="set logging level to DEBUG",
)
def version(
    config: click.Path | None,
    path_repo: click.Path,
    version_scheme: str | None,
    local_scheme: str | None,
    version_file: click.Path | None,
    version_file_template: str | None,
    relative_to: click.Path | None,
    tag_regex: str | None,
    parentdir_prefix_version: str | None,
    fallback_version: str | None,
    next: bool | None,  # noqa: A002
    no_print: bool | None,
    no_version_file: bool | None,
    debug: bool | None,
) -> None:
    if debug:
        logging.basicConfig(level=logging.DEBUG)
        logging.debug("Logging level has been set to DEBUG")

    # Priority to find configuration file user configuration -> foxy-project.toml -> pyproject.toml
    config_file = (
        config or _find_file(str(path_repo), name=FOXY_PROJECT_TOML) or _find_file(str(path_repo), name=PYPROJECT)
    )

    configuration = (
        Configuration.from_file(str(config_file), root=os.path.abspath(str(path_repo)))
        if config_file is not None
        else Configuration()
    )
    # Configurations from command line override configuration from file
    configuration = configuration.copy(
        {
            "version_scheme": version_scheme,
            "local_scheme": local_scheme,
            "version_file": Path(str(version_file)) if version_file is not None else None,
            "version_file_template": version_file_template,
            "relative_to": Path(str(relative_to)) if relative_to is not None else None,
            "tag_regex": tag_regex,
            "parentdir_prefix_version": parentdir_prefix_version,
            "fallback_version": fallback_version,
        }
    )

    if next and configuration.version_scheme == "semver-conventional-commit-foxy":
        configuration.version_scheme = semver_conventional_commit_next_foxy

    if next and configuration.version_scheme == "calendar-conventional-commit-foxy":
        configuration.version_scheme = calendar_conventional_commit_foxy_next

    if next and configuration.version_scheme == "pep440-conventional-commit-foxy":
        configuration.version_scheme = pep440_conventional_commit_next_foxy

    if next:
        configuration.local_scheme = "no-local-version"

    if no_version_file:
        configuration.version_file = None

    version = get_version(
        root=configuration.root,
        version_scheme=configuration.version_scheme,
        local_scheme=configuration.local_scheme,
        version_file=configuration.version_file,
        version_file_template=configuration.version_file_template,
        relative_to=configuration.relative_to,
        tag_regex=configuration.tag_regex,
        fallback_version=configuration.fallback_version,
        normalize=configuration.normalize,
    )
    if not no_print:
        click.echo(version)


def _find_file(parent: str, name: str) -> str | None:
    for directory in _walk_potential_roots(os.path.abspath(parent)):
        pyproject = os.path.join(directory, name)
        if os.path.isfile(pyproject):
            return pyproject

    return None


def _walk_potential_roots(root: str, *, search_parents: bool = True) -> Iterator[Path]:
    """
    Iterate though a path and each of its parents.
    :param root: File path.
    :param search_parents: If ``False`` the parents are not considered.
    """
    root_path = Path(root)
    yield root_path
    if search_parents:
        yield from root_path.parents



# SPDX-FileCopyrightText: 2024-present Fabien Hermitte
#
# SPDX-License-Identifier: MIT


from __future__ import annotations

import click

from foxy_project._version import __version__
from foxy_project.changelog.changelog import changelog
from foxy_project.version.version import version


@click.group()
@click.version_option(version=__version__, prog_name="Foxy project")
def cli() -> None:
    pass


cli.add_command(changelog)
cli.add_command(version)



# SPDX-FileCopyrightText: 2016-2024 Michael Bryan <michaelfbryan@gmail.com> - Ken Mijime <kenaco666@gmail.com>
# SPDX-FileCopyrightText: 2024-present Fabien Hermitte
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from unittest.mock import MagicMock
from unittest.mock import Mock
from unittest.mock import patch

import pytest

from foxy_project.changelog.repository import GitRepository
from git import Repo
from git.objects import Commit


@patch("foxy_project.changelog.repository.Repo", autospec=True)
@patch.object(GitRepository, "_extract_release_args", return_value=("title", "date", "sha"))
@patch.object(GitRepository, "_extract_note_args", return_value=("sha", "change_type", "description"))
@patch.object(GitRepository, "_get_git_url", return_value="git@github.com:Michael-F-Bryan/auto-changelog.git")
def test_include_unreleased(mock_ggu, mock_ena, mock_era, mock_repo):
    mock_repo.return_value.iter_commits.return_value = [MagicMock(spec=Commit), MagicMock(spec=Commit)]

    repository = GitRepository(".", skip_unreleased=False)
    changelog = repository.generate_changelog()

    assert changelog.releases[0].title == "Unreleased"


@patch("foxy_project.changelog.repository.Repo", autospec=True)
@patch.object(GitRepository, "_extract_release_args", return_value=("title", "date", "sha"))
@patch.object(GitRepository, "_extract_note_args", return_value=("sha", "change_type", "description"))
@patch.object(GitRepository, "_get_git_url", return_value="git@github.com:Michael-F-Bryan/auto-changelog.git")
def test_latest_version(mock_ggu, mock_ena, mock_era, mock_repo):
    mock_repo.return_value.iter_commits.return_value = [MagicMock(spec=Commit), MagicMock(spec=Commit)]

    repository = GitRepository(".", latest_version="v1.2.3")
    changelog = repository.generate_changelog()

    assert changelog.releases[0].title == "v1.2.3"


def test_index_init():
    commit1 = Mock()
    commit2 = Mock()
    tagref1 = Mock(commit=commit1)
    tagref1.name = "1.0.0"
    tagref2 = Mock(commit=commit2)
    tagref2.name = "2.0.0"
    tagref3 = Mock(commit=commit2)
    tagref3.name = "2.0.0"

    repo_mock = Mock(spec=Repo, tags=[tagref1, tagref2, tagref3])

    # no prefix
    tag_prefix = ""
    # we are using default tag pattern => semantic versioning
    tag_pattern = None

    index = GitRepository._init_commit_tags_index(  # pylint: disable=protected-access
        repo_mock, tag_prefix, tag_pattern
    )
    assert index == {commit1: [tagref1], commit2: [tagref2, tagref3]}


@pytest.mark.parametrize("tags", [["1.0.0", "1.0.0-beta", "v2.0.0", "v3.0.0", "4.0", "5", "v6"]])
@pytest.mark.parametrize(
    "tag_prefix, tag_pattern, expected_tags",
    [
        ("", None, ["1.0.0", "1.0.0-beta"]),
        ("v", None, ["v2.0.0", "v3.0.0"]),
        ("x", None, []),
        ("", "([1-9])", ["5"]),
        ("v", "([1-9])", ["v6"]),
        ("x", "([1-9])", []),
    ],
)
def test_tag_pattern(tags, tag_prefix, tag_pattern, expected_tags):
    tag_refs = []
    for tag in tags:
        tag_ref = Mock(commit=Mock())
        tag_ref.name = tag
        tag_refs.append(tag_ref)

    repo_mock = Mock(spec=Repo, tags=tag_refs)

    selected_tag_refs = GitRepository._init_commit_tags_index(  # pylint: disable=protected-access
        repo_mock, tag_prefix, tag_pattern
    )

    selected_tags = []
    for selected_tag_ref_list in selected_tag_refs.values():
        for selected_tag_ref in selected_tag_ref_list:
            selected_tags.append(selected_tag_ref.name)

    assert sorted(selected_tags) == sorted(expected_tags)


@pytest.mark.parametrize(
    "message, expected",
    [
        ("", ("", "", "", "", "")),
        ("feat: description", ("feat", "", "description", "", "")),
        ("feat(scope): description", ("feat", "scope", "description", "", "")),
        ("feat: description\n\nbody", ("feat", "", "description", "body", "")),
        ("feat: description\n\nbody\nbody2", ("feat", "", "description", "body\nbody2", "")),
        ("feat: description\n\n", ("feat", "", "description", "", "")),
        ("feat: description\n\nbody\nbody2\nbody3", ("feat", "", "description", "body\nbody2\nbody3", "")),
        ("feat: description\n\nparagraph1\n\nparagraph2", ("feat", "", "description", "paragraph1\n\nparagraph2", "")),
        (
            "feat: description\n\nbody\n\nFooter: footer text",
            ("feat", "", "description", "body", "Footer: footer text"),
        ),
        ("feat: description\n\nbody\n\nCloses #11", ("feat", "", "description", "body", "Closes #11")),
        (
            "feat: description\n\nbody\n\nFooter #1\nFooter: second",
            ("feat", "", "description", "body", "Footer #1\nFooter: second"),
        ),
        (
            "feat: description\n\nbody1\nbody2\n\nbodypara2\n\nFooter #1\n"
            "Footer: second but this a rather long footer\nspanning accross two lines\nFooter: third",
            (
                "feat",
                "",
                "description",
                "body1\nbody2\n\nbodypara2",
                "Footer #1\nFooter: second but this a rather long footer\nspanning accross two lines\nFooter: third",
            ),
        ),
        (
            "feat: description\n\nFooter: footer text",
            ("feat", "", "description", "", "Footer: footer text"),
        ),
        ("feat: description\n\nCloses #11", ("feat", "", "description", "", "Closes #11")),
        (
            "feat: description\n\nFooter #1\nFooter: second",
            ("feat", "", "description", "", "Footer #1\nFooter: second"),
        ),
        ("feat(scope): description\n\nbody\n\nFooter #1", ("feat", "scope", "description", "body", "Footer #1")),
        (
            "fix: correct minor typos in code\n\nsee the issue for details\n\non typos fixed.\n\n"
            "Reviewed-by: Z\nRefs #133",
            (
                "fix",
                "",
                "correct minor typos in code",
                "see the issue for details\n\non typos fixed.",
                "Reviewed-by: Z\nRefs #133",
            ),
        ),
        ("feat!: description\n\nbody\n\nFooter #1", ("feat", "", "description", "body", "Footer #1")),
        ("feat(scope)!: description\n\nbody\n\nFooter #1", ("feat", "scope", "description", "body", "Footer #1")),
        (
            "feat: description\n\nbody\n\nBREAKING-CHANGE: details",
            ("feat", "", "description", "body", "BREAKING-CHANGE: details"),
        ),
        (
            "feat: description\n\nbody\n\nBREAKING CHANGE: details",
            ("feat", "", "description", "body", "BREAKING CHANGE: details"),
        ),
        ("feat: description\n\nBREAKING-CHANGE: details", ("feat", "", "description", "", "BREAKING-CHANGE: details")),
        ("feat: description\n\nBREAKING CHANGE: details", ("feat", "", "description", "", "BREAKING CHANGE: details")),
    ],
)
def test_parse_conventional_commit_with_empty_message(message, expected):
    assert expected == GitRepository._parse_conventional_commit(message)  # pylint: disable=protected-access


@patch("foxy_project.changelog.repository.Repo", autospec=True)
@patch.object(GitRepository, "_get_git_url")
def test_get_remote_url(mock_ggu, mock_repo):
    mock_ggu.return_value = "git@github.com:Michael-F-Bryan/auto-changelog.git"
    remote_url = GitRepository(".")._remote_url(remote="origin")  # pylint: disable=protected-access
    expected = "https://github.com/Michael-F-Bryan/auto-changelog"
    assert expected == remote_url
    mock_ggu.return_value = "https://github.com/Michael-F-Bryan/auto-changelog.git"
    remote_url = GitRepository(".")._remote_url(remote="origin")  # pylint: disable=protected-access
    expected = "https://github.com/Michael-F-Bryan/auto-changelog"
    assert expected == remote_url


@patch("foxy_project.changelog.repository.Repo", autospec=True)
@patch.object(GitRepository, "_remote_url", return_value="https://github.com/Michael-F-Bryan/auto-changelog")
def test_issue_from_git_remote_url(mock_ru, mock_repo):
    remote_url = GitRepository(".")._issue_from_git_remote_url(remote="origin")  # pylint: disable=protected-access
    expected = "https://github.com/Michael-F-Bryan/auto-changelog/issues/{id}"
    assert expected == remote_url


@pytest.mark.parametrize(
    "input, expected",
    [
        ("git@github.com:Michael-F-Bryan/auto-changelog.git", "https://github.com/Michael-F-Bryan/auto-changelog"),
        ("https://github.com/Michael-F-Bryan/auto-changelog.git", "https://github.com/Michael-F-Bryan/auto-changelog"),
        ("git@gitlab.com:Michael-F-Bryan/auto-changelog.git", "https://gitlab.com/Michael-F-Bryan/auto-changelog"),
        ("https://gitlab.com/Michael-F-Bryan/auto-changelog.git", "https://gitlab.com/Michael-F-Bryan/auto-changelog"),
        (
            "https://username:0123456789abcdef0123456789abcdef01234567@github.com/Michael-F-Bryan/auto-changelog.git",
            "https://github.com/Michael-F-Bryan/auto-changelog",
        ),
        (
            "https://0123456789abcdef0123456789abcdef01234567@github.com/Michael-F-Bryan/auto-changelog.git",
            "https://github.com/Michael-F-Bryan/auto-changelog",
        ),
        (
            "https://username:0123456789abcdef0123456789abcdef01234567@gitlab.com/Michael-F-Bryan/auto-changelog.git",
            "https://gitlab.com/Michael-F-Bryan/auto-changelog",
        ),
    ],
)
def test_remote_url(input, expected):  # pylint: disable=redefined-builtin
    assert expected == GitRepository._sanitize_remote_url(input)  # pylint: disable=protected-access



# SPDX-FileCopyrightText: 2016-2024 Michael Bryan <michaelfbryan@gmail.com> - Ken Mijime <kenaco666@gmail.com>
# SPDX-FileCopyrightText: 2024-present Fabien Hermitte
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import logging
import re

from abc import ABC
from abc import abstractmethod
from enum import Enum
from typing import Any
from typing import Callable

from foxy_project.changelog import default_issue_pattern


# Default aim for Semver tags.
# Original Semver source: https://semver.org/#is-there-a-suggested-regular-expression-regex-to-check-a-semver-string
semver_nammed_regex = r"(?P<version>((?P<major>0|[1-9]\d*)\.(?P<minor>0|[1-9]\d*)\.(?P<patch>0|[1-9]\d*))(?:-(?P<prerelease>(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\+(?P<buildmetadata>[0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?)"  # noqa: E501
calendar_nammed_regex = r"(?P<version>((?P<major>\d{4})\.(?P<minor>\d{2})\.(?P<patch>[1-9]\d*))(?:-(?P<prerelease>(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\+(?P<buildmetadata>[0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?)"  # noqa: E501
semver_regex = r"((?:0|[1-9]\d*\.(?:0|[1-9]\d*)\.(?:0|[1-9]\d*)(?:-(?:(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\+([0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?))"  # noqa: E501


class ChangeType(Enum):
    BUILD = "build"
    CI = "ci"
    CHORE = "chore"
    DEPS = "deps"
    DOCS = "docs"
    FEAT = "feat"
    FIX = "fix"
    PERF = "perf"
    REFACTOR = "refactor"
    REVERT = "revert"
    STYLE = "style"
    TEST = "test"
    TOOLS = "tools"
    VERSION = "version"


class Note:
    def __init__(
        self,
        sha: str,
        change_type: ChangeType | str,
        description: str,
        scope: str = "",
        body: str = "",
        footer: str = "",
    ):
        self.sha = sha
        self.change_type = ChangeType(change_type) if change_type else change_type  # TODO Hmm..
        self.scope = scope
        self.description = description
        self.body = body
        self.footer = footer

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Note):
            return NotImplemented
        return (
            self.sha == other.sha
            and self.change_type == other.change_type
            and self.description == other.description
            and self.body == other.body
            and self.footer == other.footer
        )


class DependencyUpdate:
    def __init__(
        self,
        name: str,
        previous_version: str,
        next_version: str,
    ):
        self.name = name
        self.previous_version = previous_version
        self.next_version = next_version

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, DependencyUpdate):
            return NotImplemented
        return (
            self.name == other.name
            and self.previous_version == other.previous_version
            and self.next_version == other.next_version
        )


class Release(Note):
    def __init__(
        self,
        title: str,
        tag: str,
        date: str,
        sha: str,
        *args: Any,
        change_type: str = "chore",
        description: str = "",
        **kwargs: Any,
    ):
        super().__init__(sha, change_type, description, *args, **kwargs)
        self.title = title
        self.tag = tag
        self.date = date
        self._notes: list[Note] = []
        self._changes_indicators = {type_: False for type_ in ChangeType}
        self.diff_url: str | None = None
        self.previous_tag: str | None = None

    @property
    def builds(self) -> tuple[Note, ...]:
        return self._notes_with_type(ChangeType.BUILD)

    @property
    def ci(self) -> tuple[Note, ...]:
        return self._notes_with_type(ChangeType.CI)

    @property
    def chore(self) -> tuple[Note, ...]:
        return self._notes_with_type(ChangeType.CHORE)

    @property
    def deps(self) -> tuple[Note, ...]:
        return self._notes_with_type(ChangeType.DEPS)

    @property
    def deps_table(self) -> list[DependencyUpdate]:
        deps_notes = self._notes_with_type(ChangeType.DEPS)
        map_deps: dict[str, DependencyUpdate] = {}
        regex = re.compile(f"from {semver_regex} to {semver_regex}")
        for deps_note in deps_notes:
            match = regex.match(deps_note.description)
            if match:
                if deps_note.scope in map_deps:
                    map_deps[deps_note.scope].previous_version = match.group(1)
                else:
                    map_deps[deps_note.scope] = DependencyUpdate(
                        name=deps_note.scope, previous_version=match.group(1), next_version=match.group(3)
                    )

        return list(map_deps.values())

    @property
    def docs(self) -> tuple[Note, ...]:
        return self._notes_with_type(ChangeType.DOCS)

    @property
    def features(self) -> tuple[Note, ...]:
        return self._notes_with_type(ChangeType.FEAT)

    @property
    def fixes(self) -> tuple[Note, ...]:
        return self._notes_with_type(ChangeType.FIX)

    @property
    def performance_improvements(self) -> tuple[Note, ...]:
        return self._notes_with_type(ChangeType.PERF)

    @property
    def refactorings(self) -> tuple[Note, ...]:
        return self._notes_with_type(ChangeType.REFACTOR)

    @property
    def reverts(self) -> tuple[Note, ...]:
        return self._notes_with_type(ChangeType.REVERT)

    @property
    def style_changes(self) -> tuple[Note, ...]:
        return self._notes_with_type(ChangeType.STYLE)

    @property
    def tests(self) -> tuple[Note, ...]:
        return self._notes_with_type(ChangeType.TEST)

    @property
    def tools(self) -> tuple[Note, ...]:
        return self._notes_with_type(ChangeType.TOOLS)

    @property
    def version(self) -> tuple[Note, ...]:
        return self._notes_with_type(ChangeType.VERSION)

    @property
    def has_builds(self) -> bool:
        return self._has(ChangeType.BUILD)

    @property
    def has_ci(self) -> bool:
        return self._has(ChangeType.CI)

    @property
    def has_chore(self) -> bool:
        return self._has(ChangeType.CHORE)

    @property
    def has_deps(self) -> bool:
        return self._has(ChangeType.DEPS)

    @property
    def has_docs(self) -> bool:
        return self._has(ChangeType.DOCS)

    @property
    def has_features(self) -> bool:
        return self._has(ChangeType.FEAT)

    @property
    def has_fixes(self) -> bool:
        return self._has(ChangeType.FIX)

    @property
    def has_performance_improvements(self) -> bool:
        return self._has(ChangeType.PERF)

    @property
    def has_refactorings(self) -> bool:
        return self._has(ChangeType.REFACTOR)

    @property
    def has_reverts(self) -> bool:
        return self._has(ChangeType.REVERT)

    @property
    def has_style_changes(self) -> bool:
        return self._has(ChangeType.STYLE)

    @property
    def has_tests(self) -> bool:
        return self._has(ChangeType.TEST)

    @property
    def has_tools(self) -> bool:
        return self._has(ChangeType.TOOLS)

    @property
    def has_version(self) -> bool:
        return self._has(ChangeType.VERSION)

    def add_note(self, note: Note) -> None:
        self._notes.append(note)
        self._changes_indicators[note.change_type] = True

    def set_compare_url(self, diff_url: str, previous_tag: str) -> None:
        self.previous_tag = previous_tag
        self.diff_url = diff_url.format(previous=previous_tag, current=self.tag)

    def _notes_with(self, predicate: Callable) -> tuple[Note, ...]:
        filtered_note = filter(predicate, self._notes)
        return tuple(sorted(filtered_note, key=lambda note: note.scope))

    def _notes_with_type(self, type_: ChangeType) -> tuple[Note, ...]:
        return self._notes_with(lambda x: x.change_type == type_)

    def _has(self, type_: ChangeType) -> bool:
        return self._changes_indicators[type_]


class Changelog:  # pylint: disable=too-many-instance-attributes
    def __init__(  # pylint: disable=too-many-arguments
        self,
        title: str = "Changelog",
        description: str = "",
        issue_pattern: str | None = None,
        issue_url: str | None = None,
        tag_prefix: str = "",
        tag_pattern: str | None = None,
    ):
        self.title = title
        self.description = description
        logging.debug(default_issue_pattern)
        self.issue_pattern = issue_pattern or default_issue_pattern
        logging.debug(self.issue_pattern)
        self.issue_url = issue_url or ""
        logging.debug(issue_url)
        logging.debug(self.issue_url)
        self.tag_prefix = tag_prefix
        self.tag_pattern = tag_pattern or semver_nammed_regex
        self._releases: list[Release] = []
        self._current_release: Release | None = None

    def add_release(self, *args, **kwargs):
        """Add new Release. Require same arguments as :class:`Release`"""
        release = Release(*args, **kwargs)
        self._releases.append(release)
        self._current_release = release

    def add_note(self, *args, **kwargs):
        """Add new Note to current release. Require same arguments as :class:`Note`"""
        try:
            note = Note(*args, **kwargs)
        except ValueError as err:
            # Ignore exceptions raised by unsupported commit type.
            locallogger = logging.getLogger("Changelog.add_note")
            locallogger.debug("Ignore exception raised by unsupported commit: %s", err)
            return

        if not self._current_release:
            raise ValueError("There is no release, note can be added to")
        self._current_release.add_note(note)

    @property
    def releases(self) -> tuple[Release, ...]:
        """Returns iterable of releases sorted by date (newer first)"""
        return tuple(sorted(self._releases, key=lambda r: r.date, reverse=True))


class RepositoryInterface(ABC):  # pylint: disable=too-few-public-methods
    @abstractmethod
    def generate_changelog(  # pylint: disable=too-many-arguments
        self,
        title: str,
        description: str,
        remote: str,
        issue_pattern: str | None,
        issue_url: str | None,
        diff_url: str | None,
        starting_commit: str,
        stopping_commit: str,
    ) -> Changelog:
        raise NotImplementedError


class PresenterInterface(ABC):  # pylint: disable=too-few-public-methods
    @abstractmethod
    def present(self, changelog: Changelog) -> str:
        raise NotImplementedError



# SPDX-FileCopyrightText: 2010-2024 Ronny Pfannschmidt <opensource@ronnypfannschmidt.de>
# SPDX-FileCopyrightText: 2024-present Fabien Hermitte
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Any

from foxy_project.config.pyproject_reading import PYPROJECT
from foxy_project.config.pyproject_reading import get_args_for_pyproject
from foxy_project.config.pyproject_reading import read_changelog_config


@dataclass
class Configuration:
    """Global configuration model"""

    gitlab: bool = False
    github: bool = True
    title: str | None = "Changelog"
    description: str | None = None
    output: Path = Path("CHANGELOG.md")
    remote: str = "origin"
    latest_version: str | None = None
    unreleased: bool = False
    template: str = "compact"
    diff_url: str | None = None
    issue_url: str | None = None
    issue_pattern: str = r"(#([\w-]+))"
    tag_pattern: str = "semver"
    tag_prefix: str = ""
    stdout: bool = False
    starting_commit: str = ""
    stopping_commit: str = "HEAD"
    dist_name: str | None = None
    root: str = "."

    @classmethod
    def from_file(
        cls,
        name: str,
        dist_name: str | None = None,
        *,
        _require_section: bool = False,
        **kwargs: Any,
    ) -> Configuration:
        """
        Read Configuration from file.
        Raises exceptions when file is not found or toml is not installed or the file has invalid format.
        """

        use_tool_name = PYPROJECT in name
        pyproject_data = read_changelog_config(
            setion_name="changelog", path=Path(name), use_tool_name=use_tool_name, require_section=_require_section
        )
        args = get_args_for_pyproject(pyproject_data, dist_name, kwargs)
        # Take configuration from project as default configuration
        if PYPROJECT in name:
            if "title" not in args:
                args["title"] = pyproject_data.project_name
            if "description" not in args:
                args["description"] = pyproject_data.project_description
        return cls.from_data(data=args)

    @classmethod
    def from_data(cls, data: dict[str, Any]) -> Configuration:
        """
        Given configuration data create a config instance.
        """
        return cls(**data)

    def copy(self, new: dict[str, Any]) -> Configuration:
        """
        Returns a copy of the current configuration updated with the no None fields of the given configuration.
        """
        self_as_dict = self.__dict__.copy()
        other_as_dict = {key: value for key, value in new.items() if value is not None}
        self_as_dict.update((key, value) for key, value in other_as_dict.items())
        return Configuration(**self_as_dict)



# SPDX-FileCopyrightText: 2010-2024 Ronny Pfannschmidt <opensource@ronnypfannschmidt.de>
# SPDX-FileCopyrightText: 2024-present Fabien Hermitte
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import re

from dataclasses import dataclass
from pathlib import Path
from re import Pattern
from typing import TYPE_CHECKING
from typing import Any
from typing import Callable
from typing import TypeAlias
from typing import Union

from foxy_project.config.pyproject_reading import PYPROJECT
from foxy_project.config.pyproject_reading import get_args_for_pyproject
from foxy_project.config.pyproject_reading import read_changelog_config


if TYPE_CHECKING:
    import os

    import setuptools_scm


VERSION_SCHEME: TypeAlias = Union[str, Callable[["setuptools_scm.version.ScmVersion"], str]]
SCMVERSION: TypeAlias = "setuptools_scm.version.ScmVersion"
PathT: TypeAlias = Union["os.PathLike[str]", str]

DEFAULT_VERSION_SCHEME = "pep440-conventional-commit-foxy"
DEFAULT_LOCAL_SCHEME = "node-and-date"

# default tag regex that tries to match PEP440 style versions with prefix consisting of dashed words
DEFAULT_TAG_REGEX = re.compile(r"^(?:[\w-]+-)?(?P<version>[vV]?\d+(?:\.\d+){0,2}[^\+]*)(?:\+.*)?$")


@dataclass
class Configuration:
    """Global configuration model"""

    version_scheme: VERSION_SCHEME = DEFAULT_VERSION_SCHEME
    local_scheme: VERSION_SCHEME = DEFAULT_LOCAL_SCHEME
    version_file: PathT | None = None
    version_file_template: str | None = None
    relative_to: PathT | None = None
    tag_regex: str | Pattern[str] = DEFAULT_TAG_REGEX
    parentdir_prefix_version: str | None = None
    fallback_version: str | None = None
    fallback_root: PathT = "."
    normalize: bool = True
    dist_name: str | None = None
    root: str = "."

    @classmethod
    def from_file(
        cls,
        name: str,
        dist_name: str | None = None,
        *,
        _require_section: bool = False,
        **kwargs: Any,
    ) -> Configuration:
        """
        Read Configuration from file.
        Raises exceptions when file is not found or toml is not installed or the file has invalid format.
        """

        use_tool_name = PYPROJECT in name
        pyproject_data = read_changelog_config(
            setion_name="version", path=Path(name), use_tool_name=use_tool_name, require_section=_require_section
        )
        args = get_args_for_pyproject(pyproject_data, dist_name, kwargs)
        relative_to = args.pop("relative_to", name)
        return cls.from_data(relative_to=relative_to, data=args)

    @classmethod
    def from_data(cls, relative_to: str, data: dict[str, Any]) -> Configuration:
        """
        Given configuration data create a config instance.
        """
        return cls(relative_to=relative_to, **data)

    def copy(self, new: dict[str, Any]) -> Configuration:
        """
        Returns a copy of the current configuration updated with the no None fields of the given configuration.
        """
        self_as_dict = self.__dict__.copy()
        other_as_dict = {key: value for key, value in new.items() if value is not None}
        self_as_dict.update((key, value) for key, value in other_as_dict.items())
        return Configuration(**self_as_dict)



# SPDX-FileCopyrightText: 2016-2024 Michael Bryan <michaelfbryan@gmail.com> - Ken Mijime <kenaco666@gmail.com>
# SPDX-FileCopyrightText: 2024-present Fabien Hermitte
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from datetime import date
from datetime import timedelta

import pytest

from foxy_project.changelog.domain_model import Changelog


def test_empty_changelog():
    changelog = Changelog()
    assert changelog.title == "Changelog"
    assert changelog.description == ""
    assert changelog.releases == tuple()


def test_changelog_add_note_without_release():
    changelog = Changelog()
    with pytest.raises(ValueError):
        changelog.add_note(sha="", change_type="feat", description="")


def test_changelog_add_note():
    changelog = Changelog()
    changelog.add_release(title="Unreleased", tag="HEAD", date=date.today(), sha="")
    changelog.add_note(sha="123", change_type="fix", description="")
    changelog.add_note(sha="345", change_type="feat", description="")
    # unsupported_type should be ignored
    changelog.add_note(sha="567", change_type="unsupported_type", description="")

    releases = changelog.releases
    assert len(releases) == 1
    assert releases[0].title == "Unreleased"
    assert releases[0].tag == "HEAD"
    assert len(releases[0].fixes) == 1
    assert releases[0].fixes[0].sha == "123"
    assert len(releases[0].features) == 1
    assert releases[0].features[0].sha == "345"
    assert len(releases[0]._notes) == 2  # pylint: disable=protected-access


def test_changelog_sorted_releases():
    changelog = Changelog()
    changelog.add_release(title="1.2.3", tag="1.2.3", date=date.today(), sha="123")
    changelog.add_release(title="1.1.1", tag="1.1.1", date=date.today() - timedelta(days=1), sha="456")
    releases = changelog.releases

    assert len(releases) == 2
    assert releases[0].title == "1.2.3"
    assert releases[1].title == "1.1.1"
    assert releases[0].tag == "1.2.3"
    assert releases[1].tag == "1.1.1"
